package require Tk
package provide keyrepeat 0.1

namespace eval keyrepeat {  
  variable presscode NULL
  variable keydown
  set keydown(Up) 0
  set scriptdir [file dirname [info script]]
  variable ns [namespace current]

  switch [tk windowingsystem] {
    aqua {
          interp alias {} ${ns}::handleKeyPress {} ${ns}::handleKeyPressAqua
          interp alias {} ${ns}::handleKeyRelease {} ${ns}::handleKeyReleaseAqua

         }

    x11  {
          if {$tcl_platform(os)=="Linux"} {
            # on Linux, load the extension to disable autorepeat
            load [file join $scriptdir "tdar.so"]
	    tdar start
            interp alias {} ${ns}::handleKeyPress {} ${ns}::handleKeyPressTdar
            interp alias {} ${ns}::handleKeyRelease {} ${ns}::handleKeyReleaseTdar
	  } else {
	    # other X11 - compare serial codes like on Aqua
            interp alias {} ${ns}::handleKeyPress {} ${ns}::handleKeyPressX11
            interp alias {} ${ns}::handleKeyRelease {} ${ns}::handleKeyReleaseX11
	  }  
         }
 
    win32 {
          interp alias {} ${ns}::handleKeyPress {} ${ns}::handleKeyPressWin
          interp alias {} ${ns}::handleKeyRelease {} ${ns}::handleKeyReleaseWin        
          }
    default {
          error "Don't know how to handle platform [tk windowingsystem]"
          }
  }
  proc enable {w} {  
       variable ns
       # make bindings for the window w
       bind $w <KeyPress> [list ${ns}::handleKeyPress %W %K %# %t %x %y %X %Y]
       bind $w <KeyRelease> [list ${ns}::handleKeyRelease %W %K %# %t %x %y %X %Y]
       }

  # on Aqua, we have symmetric KeyPress/KeyRelease events
  # which is ugly - we need to compare serial codes
  # see http://wiki.tcl.tk/20299
  proc handleKeyPressAqua {w keysym serial stime x y rootx rooty} {
      variable keydown
      variable presscode [list $serial $stime]
      if {![info exists keydown($keysym)] || !$keydown($keysym)} {
         set keydown($keysym) 1
         event generate $w "<<PhysPress-$keysym>>" -x $x -y $y -rootx $rootx -rooty $rooty -time $stime -serial $serial
      } else {
         # only virtual press event -- ignore
      }
  }

   proc handleKeyReleaseAqua {w keysym serial stime x y rootx rooty} {
      variable keydown
      set releasecode [list $serial $stime]
      variable ns
      if {[info exists keydown($keysym)] && $keydown($keysym)} {
        # we need to check the serialcode *after* the next
	# keydown has been processed - grrrr   
	set options [list -x $x -y $y -rootx $rootx -rooty $rooty -time $stime -serial $serial]
	after 0 [list ${ns}::handleReleaseAfterPress $keysym $releasecode $w $options] 
      }	
  }

 # on Windows, the repeated key only generates a Press event
 # very easy - don't accept Press when the key is already down
 proc handleKeyPressWin {w keysym serial stime x y rootx rooty} {
      variable keydown
      if {![info exists keydown($keysym)] || !$keydown($keysym)} {
         set keydown($keysym) 1
         event generate $w "<<PhysPress-$keysym>>" -x $x -y $y -rootx $rootx -rooty $rooty -time $stime -serial $serial
      } else {
         # only virtual press event -- ignore
      }
  }

  proc handleKeyReleaseWin {w keysym serial stime x y rootx rooty} {
      variable keydown
      # very easy -- Release is always a true event
      # Ignore only, when the key is not down (??)
      # which means, the Press event arrived at another window 
      # due to focus issues
      if {[info exists keydown($keysym)] && $keydown($keysym)} {
        set keydown($keysym) 0
	event generate $w "<<PhysRelease-$keysym>>" -x $x -y $y -rootx $rootx -rooty $rooty -time $stime -serial $serial
      }
  }

 # on X11, when XkbSetDetectableAutoRepeat is set, the situation is like 
 # windows.
 # the repeated key only generates a Press event
 # very easy - don't accept Press when the key is already down
 proc handleKeyPressTdar {w keysym serial stime x y rootx rooty} {
      variable keydown
      if {![info exists keydown($keysym)] || !$keydown($keysym)} {
         set keydown($keysym) 1
         event generate $w "<<PhysPress-$keysym>>" -x $x -y $y -rootx $rootx -rooty $rooty -time $stime -serial $serial
      } else {
         # only virtual press event -- ignore
      }
  }

  proc handleKeyReleaseTdar {w keysym serial stime x y rootx rooty} {
      variable keydown
      # very easy -- Release is always a true event
      # Ignore only, when the key is not down (??)
      # which means, the Press event arrived at another window 
      # due to focus issues
      if {[info exists keydown($keysym)] && $keydown($keysym)} {
        set keydown($keysym) 0
	event generate $w "<<PhysRelease-$keysym>>" -x $x -y $y -rootx $rootx -rooty $rooty -time $stime -serial $serial
      }
  }

  # on X11 without tdar, we have symmetric KeyPress/KeyRelease events
  # which is ugly - we need to compare serial codes
  # see http://wiki.tcl.tk/20299
  # there are timing problems :-(
   proc handleKeyPressX11 {w keysym serial stime x y rootx rooty} {
      variable keydown
      variable presscode [list $serial $stime]
      if {![info exists keydown($keysym)] || !$keydown($keysym)} {
         set keydown($keysym) 1
         event generate $w "<<PhysPress-$keysym>>" -x $x -y $y -rootx $rootx -rooty $rooty -time $stime -serial $serial
      } else {
         # only virtual press event -- ignore
      }
  }

   proc handleKeyReleaseX11 {w keysym serial stime x y rootx rooty} {
      variable keydown
      set releasecode [list $serial $stime]
      variable ns
      if {[info exists keydown($keysym)] && $keydown($keysym)} {
        # we need to check the serialcode *after* the next
	# keydown has been processed - grrrr   
	set options [list -x $x -y $y -rootx $rootx -rooty $rooty -time $stime -serial $serial]
	after 0 [list ${ns}::handleReleaseAfterPress $keysym $releasecode $w $options] 
      }	
  }

  proc handleReleaseAfterPress {keysym releasecode w options} {
     variable presscode
     variable keydown
     lassign $presscode pserial ptime
     lassign $releasecode rserial rtime
     if {$pserial != $rserial || $rtime-$ptime>2} {
       # allow for a short time difference
       set keydown($keysym) 0
       #puts "Presscode $presscode Releasecode $releasecode"
       event generate $w <<PhysRelease-$keysym>> {*}$options
     }

  }
  
}
