# place your package dependencies here
package require Tk
package require Iwidgets
package require tcl3d 0.4.0
package require keyrepeat

if [catch {package require snack}] {
  set soundavailable 0
} else {
  set soundavailable 1
}


set scriptdir [file dirname [info script]]

variable degree [expr {3.14159265358979/180}]

# special packages can be placed in the profile
# directory under lib 

# define only procs hereafter
# 
proc Init {root nspace} {
  # root is the widget pathname for a ttk::frame
  # where one should place the GUI

  # ns is the namespace, where pluginmain is evaluated
  # each plugin gets its own distinct namespace
  variable ns 		$nspace
  
  # Camera position
  variable camx 	0.0
  variable camy 	-50.0
  variable camz 	50.0
  # Perspective angle in y direction ~ focal length
  variable alpha_GL	38.0
  # Camera direction
  variable turnangle    0.0
  variable tiltangle    45.0
  # Object rotation around the z-Axis
  variable objangle	0.0
  
  # the size of a pixel in mm
  variable px 0.3697

  variable flyspeed 	5 ;# mm/s
  variable runspeed 	10 ;# mm/s
  variable rundirection NULL
  variable flydirection NULL
  variable turnomega 	0 ;# degree/s
  variable lasttime	[clock clicks -milliseconds]
  
  
  variable frTogl
  variable wheelscale  	1.0
  variable toglwin


  variable cx 		0
  variable cy 		0
  variable quaking 	0
  variable eyeheight 	1.0
  variable horizon 	0
  variable mousemode 	rotate
  variable rightbutton
  variable recording	0
  variable quakemode	0
  variable warping	0
  variable showeye	0

  # cut circle from selection
  variable selectionmode circle
  variable selectionmask {}

  # temporary float object
  variable bfl		[bildfloat]
  variable mask		[bild16]
  variable pickray	[tcl3dVector GLfloat 1]

  #sound?
  variable soundavailable
  variable scriptdir

  # contour lines
  variable contour_dist 1.0
  variable contour_halfwidth 0.03
   
  # recorded positions for video export
  variable viewlist {}

    variable teapotscale 0.2
    variable teapott  {0.1 -0.07 0.32}
    variable teapotrot {0 -100 -30}
    variable light_position { -50.0 50.0 50.0 0.0 }
  
  set frTogl [frame $root.togl]
  set frBttn [frame $root.bttn]
  set frInfo [frame $root.info]

  grid $frTogl -row 0 -column 0 -sticky news
  grid $frBttn -row 1 -column 0 -sticky nws 
  grid $frInfo -row 2 -column 0 -sticky news
  grid rowconfigure $root 0 -weight 1
  grid columnconfigure $root 0 -weight 1

  set toglwin [togl $frTogl.toglwin -double true -depth true \
                       -displayproc ${ns}::tclDisplayFunc \
                       -reshapeproc ${ns}::tclReshapeFunc \
                       -createproc  ${ns}::tclCreateFunc]
  pack $toglwin -side top -expand 1 -fill both

  ttk::button $frBttn.b -text "Build" \
         -command ${ns}::make3d

  ttk::button $frBttn.exportview -text "Export PovRay" -command ${ns}::PovRayExportButton
  ttk::button $frBttn.options -text "Options" -command ${ns}::optiondialog
  ttk::checkbutton $frBttn.record -text "Record" -variable ${ns}::recording -command ${ns}::RecordStartStop -takefocus 0
  ttk::checkbutton $frBttn.quake -text "Quake" -variable ${ns}::quakemode -command ${ns}::ToggleMouseMode -takefocus 0
  ttk::checkbutton $frBttn.eye -text "Eye" -variable ${ns}::showeye -command ${ns}::ToggleEye -takefocus 0 
  ttk::label $frBttn.eyelabel -text "Horizon"
  scale $frBttn.eyescale -orient horizontal -variable ${ns}::horizon -from -20.0 -to 20.0 \
     -resolution 0 -showvalue false -command [list ${ns}::updateView] -takefocus 0

  pack {*}[winfo children $frBttn] -side left -anchor w -expand 1

  set glInfo "OpenGL [glGetString GL_RENDERER] Version [glGetString GL_VERSION]"
  ttk::label $frInfo.l2 -text $glInfo
  pack {*}[winfo children $frInfo] -pady 2 -side top -expand 1 -fill x

  keyrepeat::enable $toglwin
  # enable virtual events for single keypress/release 
  # without autorepeat
  
  if {[tk windowingsystem]=="aqua"} {
    set rightbutton <2>
  } else {
    set rightbutton <3>
  }


  # binding for mousemode=rotate
  MouseMode rotate

  # binding for the mousewheel
  # the wheel means always going forward/backward
  bind $toglwin <MouseWheel> [list ${ns}::handleWheel %D]
  # on X11, buttons 4&5 are mousewheel events

  if {[tk windowingsystem] == "x11"} {
    bind $toglwin <4> [list ${ns}::handleWheel +1]
    bind $toglwin <5> [list ${ns}::handleWheel -1]
  }
  
  if {[tk windowingsystem] == "win32"} {
    # the mouse wheel on Windows reports 
    # an increment of 120, while X11 and Mac 
    # report only 1 per click
    set wheelscale [expr {$wheelscale/120.0}]
  }

  makegrid
  # make floor

  # initialize sound
  if {$soundavailable} {
    snack::sound ${ns}::wumms -file [file join $scriptdir rocklf1a.wav]
  }
  # the return value of this proc is used as the 
  # tab name. If not given, XRayLab uses the directory name
  return "XRay3D"
}

proc ToggleMouseMode {} {
  variable quakemode
  variable tiltangle
  if {$quakemode} {
    MouseMode quake
    set tiltangle 0
  } else {
    MouseMode rotate
  }  
  advance [UnitVectorLook] 0
}

proc ToggleEye  {} {
  variable showeye
  if {$showeye} {indicator show} else {indicator hide}
}

proc FlightControlBindings  {} {
  variable toglwin
  variable ns

  bind $toglwin <<PhysPress-Up>> [list ${ns}::StartFlight forward %x %y]
  bind $toglwin <<PhysRelease-Up>> [list ${ns}::StopFlight]
  
  bind $toglwin <<PhysPress-Down>> [list ${ns}::StartFlight backward %x %y]
  bind $toglwin <<PhysRelease-Down>> [list ${ns}::StopFlight]
  
  bind $toglwin <<PhysPress-Right>> [list ${ns}::StartFlight right %x %y]
  bind $toglwin <<PhysRelease-Right>> [list ${ns}::StopFlight]
  
  bind $toglwin <<PhysPress-Left>> [list ${ns}::StartFlight left %x %y]
  bind $toglwin <<PhysRelease-Left>> [list ${ns}::StopFlight]
}

proc MouseMode {mode} {
  variable mousemode
  variable modestack
  variable rightbutton
  variable toglwin
  variable ns
  


  switch $mode {
    rotate {
      # manipulate the object
      bind $toglwin <1> [list ${ns}::StartRot %x %y]
      bind $toglwin <B1-Motion> [list ${ns}::handleRotate %x %y]
      bind $toglwin <Motion> {}
      FlightControlBindings

      $toglwin configure -cursor {}
      # remove bindings for wasd game control
      foreach key {w a s d} {
       bind $toglwin <<PhysPress-$key>> {}
       bind $toglwin <<PhysRelease-$key>> {}
      }
      bind $toglwin <Leave> {}
    }

    quake {
      # walk around the object, similar to 1st person shooter
      # 
      InitQuake
      bind $toglwin <1> [list ${ns}::ToggleQuake %x %y]
      bind $toglwin <Motion> [list ${ns}::handleQuakeTurn %x %y %W]
      bind $toglwin <B1-Motion> {}
      FlightControlBindings

      bind $toglwin <Leave> [list ${ns}::handleQuakeLeave %x %y]

      bind $toglwin <<PhysPress-w>> [list ${ns}::StartRun forward %x %y]
      bind $toglwin <<PhysRelease-w>> ${ns}::StopRun
      bind $toglwin <<PhysPress-s>> [list ${ns}::StartRun backward %x %y]
      bind $toglwin <<PhysRelease-s>> ${ns}::StopRun
      bind $toglwin <<PhysPress-a>> [list ${ns}::StartRun left %x %y]
      bind $toglwin <<PhysRelease-a>> ${ns}::StopRun
      bind $toglwin <<PhysPress-d>> [list ${ns}::StartRun right %x %y]
      bind $toglwin <<PhysRelease-d>> ${ns}::StopRun

      $toglwin configure -cursor {}
      bind $toglwin $rightbutton ${ns}::Shoot
    }

    fly {
      # fly through the object 
      # triggered by the arrow keys in quake mode
      bind $toglwin <Motion> [list ${ns}::handleFlightTurn %x %y %W]
      bind $toglwin <1> {}
      $toglwin configure -cursor none
      bind $toglwin <Leave> {}
      bind $toglwin $rightbutton ${ns}::Shoot
    }

    push {
       lappend modestack $mousemode
       # for push and pop, don't change
       # the mode variable
       return
    }

    pop {
       set oldmode [lindex $modestack end]
       set modestack [lrange modestack 0 end-1]
       MouseMode $oldmode
       # for push and pop, don't change
       # the mode variable
       return
    }

    default {
      error "Mouse mode must be rotate, quake or fly"
    }
  }
  
  set mousemode $mode
}

if {[tk windowingsystem]!="aqua"} {
  variable warping_available 1
  proc mousecenter {x y} {
    variable warping
    variable toglwin
    variable warpx
    variable warpy
    variable cx
    variable cy
  
    set warping 1
    set warpx [expr {[winfo width $toglwin]/2}]
    set warpy [expr {[winfo height $toglwin]/2}]
    event generate $toglwin <Motion> -warp 1 -x $warpx -y $warpy
    set warping 0
    set cx $warpx
    set cy $warpy
  }
} else {
  # on Aqua, warping doesn't work :(
  variable warping_available 0
  proc mousecenter {x y} {
    variable cx
    variable cy
    set cx $x
    set cy $y
  }
}  

proc InitQuake {} {
  variable quaking 0
}

proc ToggleQuake {x y} {
  variable quaking
  variable warping_available
  variable toglwin
  if {$quaking} {
    set quaking 0
    $toglwin configure -cursor {}
  } else {
    set quaking 1
    if {$warping_available} {
      # hide mouse cursor
       $toglwin configure -cursor none
    }
    mousecenter $x $y
    focus $toglwin
  }  
}

proc handleQuakeLeave {x y} {
  variable toglwin
  variable quaking
  if {$quaking}  {
    mousecenter $x $y
  }  
}

proc StartRun {direction x y} {
  variable rundirection $direction
  variable lasttime [clock clicks -milliseconds]
  variable quaking
  if {!$quaking} {ToggleQuake $x $y}
  updateView
}

proc StopRun {} {
  variable rundirection NULL
}

proc StartRot {x y} {
  variable cx
  variable cy
  variable toglwin
  set cx $x
  set cy $y
  
  # On Win32, it is necessary to set the focus 
  # to recieve a MouseWheel event
  focus $toglwin 
}

proc handleRotate {x y} {
  variable cx 
  variable cy
  variable turnangle
  variable tiltangle
  variable objangle
  variable camx
  variable camy
  variable camz
  variable toglwin

  # 2D x-movement rotates the object around the z-axis
  set objangle [expr {$objangle + 180 * (double($x - $cx) / [winfo height $toglwin])}]
  set cx $x
  
  # 2D y-movement tilts the camera and makes it orbit around (0,0)
  #
  set tiltangledelta [expr {180 * (double($y - $cy) / [winfo height $toglwin])}]
  set tiltangle [expr {$tiltangle + $tiltangledelta}]
  set cy $y

  # compute new camera position
  set axis [UnitVectorRight]
  set newpos [RotateVector [list $camx $camy $camz] $axis [expr {-$tiltangledelta}]]
  lassign $newpos camx camy camz
  
  # redraw
  updateView 
}   

proc updateView {args} {
  variable toglwin
  indicator update
  $toglwin postredisplay
}

proc handleQuakeTurn {x y win} {
  variable quaking
  variable warping
  variable cx 
  variable cy
  variable turnangle
  variable tiltangle
  
  if {$quaking} {
    if {$warping} {
      return
    }  
    set deltax [expr {180 * (double($x - $cx) / [winfo height $win])}]
    set deltay [expr {180 * (double($y - $cy) / [winfo height $win])}]
  
    # change the direction of the view
    set tiltangle [expr {$tiltangle + $deltay}]
    set turnangle [expr {fmod($turnangle + $deltax,360)}]
    # reset increment
    set cx $x
    set cy $y
    updateView
  }
}


proc handleFlightTurn {x y win} {
  variable cx 
  variable cy
  variable turnomega
  variable tiltangle
  variable warping

  if {$warping} {
    return
  }

  set deltax [expr {180 * (double($x - $cx) / [winfo height $win])}]
  set deltay [expr {180 * (double($y - $cy) / [winfo height $win])}]

  set turnomega [expr {$deltax/2.0}]
  set tiltangle [expr {$tiltangle + $deltay}]
  # reset increment
  set cy $y
  
  # redraw
  updateView
}


proc handleWheel {dist} {
  variable camz
  variable wheelscale
  variable mousemode
  variable flyspeed
  switch $mousemode {
    # In the static modes "rotate" and "quake" , the wheel
    # moves forward/backward
    # In the fly mode, the wheel increases/decreases speed
    #
    quake {
      advance [UnitVectorLook] [expr {$dist*$wheelscale}]
      }
    rotate {
      advance [UnitVectorLook] [expr {$dist*$wheelscale}]
      }
    fly {
      set flyspeed [expr {$flyspeed*(1+$dist*$wheelscale/20.0)}]
       }
   }    
}

proc StartFlight {direction x y} {
  variable lasttime
  variable turnomega
  variable autoreleaseserial
  variable mousemode
  variable flydirection $direction

  set lasttime [clock clicks -milliseconds]
  # go ....
  MouseMode push
  MouseMode fly
  # reset rotation coordinates
  set turnomega 0
  
  mousecenter $x $y
  puts "Go $direction!!!"
  updateView
}

proc StopFlight {} {
  MouseMode pop
  variable flydirection NULL
  puts "Quietsch !!!"
}  

proc advance {direction distance} {
  variable bfl
  variable camx
  variable camy
  variable camz
  variable objangle
  variable toglwin
  variable eyeheight

  variable mask
  variable xm
  variable ym
  variable zm
  variable zmin

  variable mousemode
  
  
  lassign $direction xn yn zn 

  set camx [expr {$camx+$xn*$distance}]
  set camy [expr {$camy+$yn*$distance}]
  set camz [expr {$camz+$zn*$distance}]
 
  ###########################
  #
  # collision check 
  # first rotate the camera vector
  # to account for the rotation by objangle

  lassign [RotateVector [list $camx $camy $camz] {0 0 1} [expr {-$objangle}]] camobjx camobjy

  # then calculate the coordinates in pixels
  variable px
  set pixx [expr {int($camobjx/$px+$xm)}]
  set pixy [expr {int($ym-$camobjy/$px)}]
  
 
  if {0<=$pixx && $pixx<=511 && 0<=$pixy && $pixy<=511 && [$mask get $pixy $pixx] } {
     # z-coordinate of the corresponding pixel
     set newz [expr {[$bfl get $pixy $pixx]+$eyeheight-$zm}]
  } else {
     set newz [expr {$zmin-$zm+$eyeheight}]
  }

  switch $mousemode {
     # restrict motion such, that it stays above the ground
     quake {
          # in quake mode, walk along the surface
          set camz $newz
          }

     rotate  { 
          # in rotate mode, allow to pass the ground
     }
         
     fly  {
          # in fly mode,  bounce off the ground
	  # 
           if {$camz<$newz} {
	     set camz $newz
	   }
          }

   }  
  
  updateView
}

proc shootmark {x y} {
  if {$x<0 || $x>512 || $y<0 || $y>512} {return}
  set x1 [expr {$x-5}] 
  set x2 [expr {$x+5}]
  set y1 [expr {$y-5}] 
  set y2 [expr {$y+5}]

  $::XRlab::can create line $x1 $y1 $x2 $y2 -fill #A04000 -tag shotmark
  $::XRlab::can create line $x1 $y2 $x2 $y1 -fill #A04000 -tag shotmark

}

variable eyevisible 0
proc indicator {cmd} {
  # show the position of the eye in the top view
  # cmd=show: update position
  # cmd=hide: hide the eye
  variable eyevisible
  variable camx
  variable camy
  variable camz

  variable objangle
  variable px
  variable xm
  variable ym

  
  set length 15 
  # draw a circle and an arrow
  switch  $cmd {
    show {
      if {$eyevisible} return
      $::XRlab::can create line 0 0 1 1  -fill green -tag {eye lookat}
      $::XRlab::can create oval 1 1 2 2  -outline green -fill white -tag {eye camera}
      set eyevisible 1
      indicator update
      }
   
   update {
      lassign [UnitVectorRight] UVRx UVRy UVRz
      lassign [RotateVector [list $camx $camy $camz] {0 0 1} [expr {-$objangle}]] camobjx camobjy
      lassign [RotateVector [UnitVectorLook] {0 0 1}  [expr {-$objangle}]] camlx camly
      set myx [expr {$camobjx/$px+$xm}]    
      set myy [expr {$ym-$camobjy/$px}] 
      set lx [expr {$myx+$length*$camlx}]
      set ly [expr {$myy-$length*$camly}]

      $::XRlab::can coords lookat $myx $myy $lx $ly 
      $::XRlab::can coords camera [expr {$myx-2}] [expr {$myy-2}] [expr {$myx+2}] [expr {$myy+2}]
     }
   hide {
      $::XRlab::can delete eye 
      set eyevisible 0
     }
   }
   
}


proc Shoot {} {
  variable shotx
  variable shoty
  variable shotz
  variable soundavailable
  variable bulletlist
  variable bulletdisplaylist
  variable toglwin


  variable xm
  variable ym
  variable zm

  variable px

  set myx [expr {$shotx/$px+$xm}]
  set myy [expr {$ym-$shoty/$px}]
  puts "$myx $myy [expr {$shotz+$zm}]"
  # 2D mark in the top view
  shootmark $myx $myy

  # 3D mark == sphere in the world
  lappend bulletlist $shotx $shoty $shotz
  glDeleteLists $bulletdisplaylist 1
  set bulletdisplaylist [glGenLists 1]
  
  glNewList $bulletdisplaylist GL_COMPILE
   foreach {sx sy sz} $bulletlist {
     glPushMatrix
      glTranslatef $sx $sy $sz
      glutSolidSphere 0.5 10 15
     glPopMatrix
   }  
  glEndList   
  $toglwin postredisplay

  if {$soundavailable} {
    wumms play
  }  
}

proc make3d {} {
    variable XR3dList
    variable bulletlist {}
    $XRlab::can delete shotmark
    
    # empty bullets on rebuild
    variable bulletdisplaylist

    if { [info exists XR3dList] } {
        glDeleteLists $XR3dList 1
    }
    
    set XR3dList [glGenLists 1]
    
    variable colortexture
    if {![info exists colortexture]} {

      set texname [tcl3dVector GLuint 1]
      glGenTextures 1 $texname
      set colortexture [$texname get 0]
    }

    glNewList $XR3dList GL_COMPILE
    
    set err [catch {buildGL} errmsg]

    glEndList

    make_colorbar
  
    if {[info exists bulletdisplaylist]} {
      glDeleteLists $bulletdisplaylist 1
    }  
    set bulletdisplaylist [glGenLists 1]
    glNewList $bulletdisplaylist GL_COMPILE
    glEndList
 

    updateView
    if {$err} { error $errmsg $::errorInfo $::errorCode }
}


proc makegrid {} {
  # coordinate on the floor 
  variable gridlist 
  variable gridtexture

  if { [info exists gridlist] } {
        glDeleteLists $gridlist 1
    }
  
  set texname [tcl3dVector GLuint 1]
  glGenTextures 1 $texname
  set gridtexture [$texname get 0]

  # work on the texture $gridtexture
  glBindTexture GL_TEXTURE_2D $gridtexture
    glEnable GL_TEXTURE_2D

    glTexParameteri GL_TEXTURE_2D GL_TEXTURE_MAG_FILTER $::GL_LINEAR
    glTexParameteri GL_TEXTURE_2D GL_TEXTURE_MIN_FILTER $::GL_LINEAR_MIPMAP_LINEAR

    glTexParameteri GL_TEXTURE_2D GL_TEXTURE_WRAP_S $::GL_REPEAT
    glTexParameteri GL_TEXTURE_2D GL_TEXTURE_WRAP_T $::GL_REPEAT
  
  
    set WIDTH  64
    set HEIGHT 64
    set teximage [tcl3dVector GLubyte [expr $WIDTH*$HEIGHT*4]]

      for { set i 0 } { $i < $HEIGHT } { incr i } {
          for { set j 0 } { $j < $WIDTH } { incr j } { 
              if {$i==$HEIGHT/2 || $j==$WIDTH/2 } {
		   # cross 
                  set value 0xff
              } else {
                  set value 0x00
              }
              $teximage set [expr {($i * $WIDTH + $j)*4 + 0}] $value
              $teximage set [expr {($i * $WIDTH + $j)*4 + 1}] $value
              $teximage set [expr {($i * $WIDTH + $j)*4 + 2}] $value
              $teximage set [expr {($i * $WIDTH + $j)*4 + 3}] 0xff
          }
      }

      gluBuild2DMipmaps GL_TEXTURE_2D 4 $WIDTH $HEIGHT \
                        GL_RGBA GL_UNSIGNED_BYTE $teximage

  set gridlist [glGenLists 1]
  glNewList $gridlist GL_COMPILE
    glBindTexture GL_TEXTURE_2D $gridtexture
    glEnable GL_TEXTURE_2D
    # hard code texturing into the display list
    glBegin GL_QUADS
       # big black quad asphalt
       glColor3f 1.0 1.0 1.0
       glTexCoord2f 0.0 0.0 ;    glVertex3f -2000 -2000 0
       glTexCoord2f 0.0 100.0 ;  glVertex3f -2000  2000 0
       glTexCoord2f 100.0 100.0 ;  glVertex3f  2000  2000 0
       glTexCoord2f 100.0 0.0 ;glVertex3f  2000 -2000 0
    glEnd
    # white grid stripes
    glDisable GL_TEXTURE_2D
  glEndList

}


proc tclCreateFunc { w } {
    variable toglwin
    variable ns
    variable scriptdir
    variable XR3dList
    variable mask
    variable bfl
    variable bulletlist {}

    glClearColor 0.8 0.8 1 0 
    glEnable GL_DEPTH_TEST
    glEnable GL_POLYGON_OFFSET_FILL
    # glEnable GL_NORMALIZE
    glShadeModel GL_SMOOTH


    $bfl read_gnuplot [file join $scriptdir pattern.gnubin]
    $XRlab::can delete shotmark
    
    # empty bullets on rebuild
    variable bulletdisplaylist

    set XR3dList [glGenLists 1]
    
    variable colortexture
    if {![info exists colortexture]} {

      set texname [tcl3dVector GLuint 1]
      glGenTextures 1 $texname
      set colortexture [$texname get 0]
    }
   
    variable zmin -7.0
    variable zmax 3.0
    variable xm 256
    variable ym 256
    variable zm 0
    
    glNewList $XR3dList GL_COMPILE
     glBindTexture GL_TEXTURE_1D $colortexture
     glTexEnvf GL_TEXTURE_ENV GL_TEXTURE_ENV_MODE $::GL_MODULATE
     set texturescale [expr {1.0/($zmax-$zmin)}]
     set textureoffset [expr {-($zmin-$zm)*$texturescale}]
     set ::currentCoeff   [list 0.0 0.0 $texturescale $textureoffset]
     glTexGeni GL_S GL_TEXTURE_GEN_MODE $::GL_OBJECT_LINEAR
     glTexGenfv GL_S $::GL_OBJECT_PLANE $::currentCoeff

     glEnable GL_TEXTURE_GEN_S
     glEnable GL_TEXTURE_1D
     glPushMatrix
       glRotatef 90.0 1.0 0.0 0.0
       glutSolidTeapot 10
     glPopMatrix
     glDisable GL_TEXTURE_GEN_S
     glDisable GL_TEXTURE_1D
    glEndList
    
  
    $mask = 0

    make_colorbar
  
    # create an empty display list for the bullets
    variable bulletdisplaylist [glGenLists 1]
    glNewList $bulletdisplaylist GL_COMPILE
    glEndList
 
}




proc tclDisplayFunc { w } {
    variable XR3dList
    variable bulletdisplaylist
    variable gridlist
    variable gridtexture
    variable colortexture
    variable ns
    variable objangle
    variable zmin
    variable zm

    variable camx 
    variable camy
    variable camz
    variable horizon
    variable pickray

    variable shotx
    variable shoty
    variable shotz

    variable degree
    variable alpha_GL

    glClear [expr $::GL_COLOR_BUFFER_BIT | $::GL_DEPTH_BUFFER_BIT]
    calcView

    glDisable GL_LIGHTING
    glPolygonOffset 0.0 0.0
    glPushMatrix
     glTranslatef 0.0 0.0 [expr {$zmin-$zm}]
     glCallList $gridlist
    glPopMatrix
    
    glPolygonOffset -1.0 -1.0
    # draw the object in front

    glPushMatrix
      variable light_position 
      

      set mat_specular { 1.0 1.0 1.0 1.0 }
      set mat_diffuse { 1.0 1.0 1.0 1.0 }
      set mat_shininess { 50.0 }
  
  
      glMaterialfv GL_FRONT GL_SPECULAR $mat_specular
      glMaterialfv GL_FRONT GL_SHININESS $mat_shininess
      glMaterialfv GL_FRONT GL_DIFFUSE $mat_diffuse
      glMaterialfv GL_FRONT GL_AMBIENT { 0.7 0.7 0.7 1.0}
      glLightfv GL_LIGHT0 GL_AMBIENT { 0.2 0.2 0.2 0}
  
      glEnable GL_LIGHT0
      glEnable GL_LIGHTING
      glLightfv GL_LIGHT0 GL_POSITION $light_position

      glRotatef $objangle 0.0 0.0 1.0

      glEnable GL_TEXTURE_1D
      glCallList $XR3dList

      glMaterialfv GL_FRONT GL_DIFFUSE { 0.1 0.1 0.1 0}
      glMaterialfv GL_FRONT GL_AMBIENT { 0.1 0.1 0.1 0}
      glMaterialfv GL_FRONT GL_SHININESS { 50 }

      glCallList $bulletdisplaylist


      # get the coordinates of the center point
      set viewport   [tcl3dOglGetViewport]
      lassign $viewport left bottom right top
      set mvmatrix_stacheln   [tcl3dOglGetDoubleState GL_MODELVIEW_MATRIX 16] 
      set projmatrix_stacheln [tcl3dOglGetDoubleState GL_PROJECTION_MATRIX 16] 
      set centerx [expr {int($left+$right)/2}]
      set centery [expr {int(($top+$bottom)/2 + ($top-$bottom)/2*tan($horizon*$degree)/tan($alpha_GL*$degree/2))}] 

      glReadPixels $centerx $centery 1 1 GL_DEPTH_COMPONENT GL_FLOAT $pickray
    glPopMatrix

    variable mousemode
    if {$mousemode!="rotate"} {
      # in quake and fly modes show navigation elements on top
      set crosslength 0.5
      set crossthick 0.002
      lassign [UnitVectorLook] lx ly lz
      lassign [UnitVectorRight] rx ry rz
      lassign [UnitVectorUp] ux uy uz
  
      set mx [expr {$camx+3*$lx}] 
      set my [expr {$camy+3*$ly}] 
      set mz [expr {$camz+3*$lz}] 
  
      # clear the depth buffer to draw the navigation on top
      glClear $::GL_DEPTH_BUFFER_BIT
      glDisable GL_LIGHTING
      
      glPushMatrix
      glTranslatef $camx $camy $camz
      glMultMatrixf [list $rx $ry $rz 0   $ux $uy $uz 0    $lx $ly $lz 0   0 0 0 1]
      glScalef $crossthick $crossthick 1
      glBegin GL_QUADS
        # draw thick black cross
        glColor3f   0 0 0
        glVertex3f -12 -3   1.2
        glVertex3f -12 3   1.2
        glVertex3f  12 3   1.2
        glVertex3f  12 -3   1.2
  
        glVertex3f -3 -12   1.2
        glVertex3f -3  12   1.2
        glVertex3f  3  12   1.2
        glVertex3f  3 -12   1.2
  
  
        # draw thin white cross
        glColor3f 1 1 1
        glVertex3f -10 -0.5   1
        glVertex3f -10  0.5   1
        glVertex3f  10  0.5   1
        glVertex3f  10 -0.5   1
  
        glVertex3f -0.5 -10   1
        glVertex3f -0.5  10   1
        glVertex3f  0.5  10   1
        glVertex3f  0.5 -10   1
  
      glEnd
      

      glPopMatrix
  
      glTranslatef $camx $camy $camz
      glMultMatrixf [list $rx $ry $rz 0   $ux $uy $uz 0    $lx $ly $lz 0   0 0 0 1]
      
      variable teapotscale
      variable teapott
      variable teapotrot
      lassign $teapotrot teapotrotx teapotroty teapotrotz
  
      glTranslatef {*}$teapott
      glRotatef $teapotrotx 1 0 0
      glRotatef $teapotroty 0 1 0
      glRotatef $teapotrotz 0 0 1
  
  
     set mat_specular { 1.0 1.0 1.0 1.0 }
     set mat_diffuse { 0.7 0.6 0.55 1.0 }
     set mat_shininess { 50.0 }
  
  
     glMaterialfv GL_FRONT GL_SPECULAR $mat_specular
     glMaterialfv GL_FRONT GL_SHININESS $mat_shininess
     glMaterialfv GL_FRONT GL_DIFFUSE $mat_diffuse
  
     glEnable GL_LIGHT0
     glEnable GL_LIGHTING
     glutSolidTeapot $teapotscale
   }

   $w swapbuffers

   # calculate the position where the cross points to
   # cross center coordinates are 0,0,1
   # lassign [gluProject 0 0 1 $mvmatrix $projmatrix $viewport] troffen centerx centery dummy
   lassign [gluUnProject $centerx $centery [$pickray get 0] $mvmatrix_stacheln $projmatrix_stacheln $viewport] peakshot shotx shoty shotz

   # control animation
   # 
   doAnimation
   
}

proc tclReshapeFunc { toglwin w h } {
    variable camx
    variable camy
    variable camz
    variable alpha_GL

    glViewport 0 0 $w $h
    glMatrixMode GL_PROJECTION
    glLoadIdentity
    gluPerspective $alpha_GL [expr double($w)/double($h)] 0.5 2000.0
    glMatrixMode GL_MODELVIEW
    calcView
    # Camera points into positive y-direction 
    # if not turned by the player

}

proc UnitVectorLook {{tilt 0} {turn 0}} {
  # calculate unit vector in the direction of the 
  # camera based on turnangle (longitude) /tiltangle (altitude)
  # tilt and turn are offsets
  variable tiltangle
  variable turnangle

  set turn [expr {$turn+$turnangle}]
  set tilt [expr {$tilt+$tiltangle}]


  variable degree 
  set sturn  [expr {sin($turn*$degree)}]
  set cturn  [expr {cos($turn*$degree)}]
  set stilt  [expr {sin($tilt*$degree)}]
  set ctilt  [expr {cos($tilt*$degree)}]
  
  set nx [expr {$sturn*$ctilt}]
  set ny [expr {$cturn*$ctilt}]
  set nz [expr {-$stilt}]
  return [list $nx $ny $nz]
}

proc UnitVectorUp {} {
  # calculate unit vector in the direction of the 
  # camera based on turnangle (longitude) /tiltangle (altitude)
  variable tiltangle
  variable turnangle
 
  set degree [expr {3.14159265358979/180}]
  set sturn  [expr {sin($turnangle*$degree)}]
  set cturn  [expr {cos($turnangle*$degree)}]
  set stilt  [expr {sin($tiltangle*$degree)}]
  set ctilt  [expr {cos($tiltangle*$degree)}]
  
  set nx [expr {$sturn*$stilt}]
  set ny [expr {$cturn*$stilt}]
  set nz [expr {$ctilt}]
  return [list $nx $ny $nz]
}

proc UnitVectorRight {} {
  # calculate unit vector in the direction of the 
  # camera based on turnangle (longitude) /tiltangle (altitude)
  variable tiltangle
  variable turnangle
 
  set degree [expr {3.14159265358979/180}]
  set sturn  [expr {sin($turnangle*$degree)}]
  set cturn  [expr {cos($turnangle*$degree)}]
  set stilt  [expr {sin($tiltangle*$degree)}]
  set ctilt  [expr {cos($tiltangle*$degree)}]
  
  set nx [expr {$cturn}]
  set ny [expr {-$sturn}]
  set nz [expr {0.0}]
  return [list $nx $ny $nz]
}

proc calcView {} {
  # helper function -- calculate parameters for gluLookAt
  # based on position and turn/tilt angle
  variable camx
  variable camy
  variable camz
  variable centerx
  variable centery
  variable centerz

  variable turnangle
  variable tiltangle
  variable objangle
  variable horizon
   
  lassign [UnitVectorLook $horizon] lx  ly  lz 
  lassign [UnitVectorUp]   upx upy upz 

  set centerx [expr {$camx+$lx}]
  set centery [expr {$camy+$ly}]
  set centerz [expr {$camz+$lz}]

  glLoadIdentity
  gluLookAt $camx $camy $camz   $centerx $centery $centerz  $upx $upy $upz
}

proc RecordStartStop {} {
  variable recording
  variable viewlist
  if {$recording} {
    # start -- empty view bucket
    set viewlist {}
  } else {
    # do nothing
    # recording is stopped automatically by the button
  }
}


proc doAnimation {} {
  variable mousemode
  variable lasttime
  variable turnangle
  variable turnomega
  variable tiltangle
  variable flyspeed
  variable runspeed
  variable rundirection
  variable flydirection
  variable ns
  
  variable recording
  variable viewlist
  variable camx
  variable camy
  variable camz
  
  variable centerx
  variable centery
  variable centerz

  variable shotx
  variable shoty
  variable shotz 

  variable objangle
  variable degree


  # Realtime engine -- measure time and update positions
  #
  set newtime [clock clicks -milliseconds]
  
  if {$recording} {
    # record the positions in the global view list for playback
    lappend viewlist $newtime $camx $camy $camz $centerx $centery $centerz $objangle
  } 

  switch $mousemode {
    fly {
      # in fly mode, go into the direction of the camera
      # and turn by turnomega
      # calculate the time which passed
      set ftime [expr {($newtime-$lasttime)/1000.0}]
      
      switch $flydirection {
	forward { 
	         set direction [UnitVectorLook]
		 set distance  [expr {$flyspeed*$ftime}]
                 set turnangle [expr {fmod($turnangle+$turnomega*$ftime,360)}]
	         }
	 backward {
	         set direction [UnitVectorLook]
		 set distance  [expr {-$flyspeed*$ftime}]
                 set turnangle [expr {fmod($turnangle-$turnomega*$ftime,360)}]
	          }
	 right  {
	         set direction [UnitVectorRight]
		 set distance  [expr {$flyspeed*$ftime}]
                 # calculate rotation speed around the point
                 # where the camera points to
		 set shotdist [expr {sqrt(($camx-$shotx)**2+($camy-$shoty)**2)}]
		 set strideangle [expr {atan2($distance,$shotdist)/$degree}]
                 set turnangle [expr {fmod($turnangle-$strideangle,360)}]
	        }
	 left   {
	         set direction [UnitVectorRight]
		 set distance  [expr {-$flyspeed*$ftime}]
                 # calculate rotation speed around the point
                 # where the camera points to
		 set shotdist [expr {sqrt(($camx-$shotx)**2+($camy-$shoty)**2)}]
		 set strideangle [expr {atan2($distance,$shotdist)/$degree}]
                 set turnangle [expr {fmod($turnangle-$strideangle,360)}]
	        }
	 NULL {return}
	default {error "flydirection is wrong ,,$flydirection''"}
      }	
      after 1 [list ${ns}::advance $direction $distance]
      set lasttime $newtime
      
    }

    quake {
      # in quake mode, go into the direction "rundirection"
      # don't turn automatically 
      # calculate the time which passed in seconds
      set ftime [expr {($newtime-$lasttime)/1000.0}]
      # calculate the direction and distance, where to go
      switch $rundirection {
         forward { 
	         set direction [UnitVectorLook]
		 set distance  [expr {$runspeed*$ftime}]
	         }
	 backward {
	         set direction [UnitVectorLook]
		 set distance  [expr {-$runspeed*$ftime}]
	          }
	 right  {
	         set direction [UnitVectorRight]
		 set distance  [expr {$runspeed*$ftime}]
	        }
	 left   {
	         set direction [UnitVectorRight]
		 set distance  [expr {-$runspeed*$ftime}]
	        }
	 NULL {return}
	 default {error "Quake direction must be forward,backward,left,right!"}
      }	 
      after 1 [list ${ns}::advance $direction $distance]
      set lasttime $newtime
      
    }

    default {
      
    }

  }

}

# 
proc popup {} {
 tk_messageBox -type ok -message "\
 3D Visualization 
 (C) 2009 Auriocus
 Click on Build to make a 3D image from the 
 currently displayed bildfloat
 "
}

if {0} {
  # gnuplot standard color palette
  proc red {cb} {
   if {$cb<0} {set cb 0}
   if {$cb>1} {set cb 1}
   expr {sqrt($cb)}
  }
  
  proc green {cb} {
   if {$cb<0} {set cb 0}
   if {$cb>1} {set cb 1}
   expr {$cb**3}
  }
  
  proc blue {cb} {
   if {$cb<0} {set cb 0}
   if {$cb>1} {set cb 1}
   set blue  [expr {sin(6.283185*$cb)}]
   if {$blue<0} {set blue 0}
   return $blue
  }
} 

# blue - green - red color palette
  proc red {cb} {
   if {$cb<0} {set cb 0}
   if {$cb>1} {set cb 1}
   return $cb
  }
  
  proc green {cb} {
   if {$cb<0} {set cb 0}
   if {$cb>1} {set cb 1}
   expr {exp(-(($cb-0.5)**2/(0.2)**2))}
  }
  
  proc blue {cb} {
   if {$cb<0} {set cb 0}
   if {$cb>1} {set cb 1}
   expr {1-$cb}
  }

# render an X-ray image with opengl
proc buildGL {} {
  variable bfl
  variable mask
  variable selectionmode
  variable selectionmask
 

  variable zmin
  variable zmax
  variable zm
  variable xm
  variable ym
  variable radius
  variable ns
  variable px
  variable colortexture

  if {[typecheck $::XRlab::imgname] != "bildfloat"} {
    error "You must display a float object for 3D!"
  }
  
  # create copy 
  $bfl = $::XRlab::imgname

  if {$::XRlab::y1zoom!=$::XRlab::y2zoom && $::XRlab::x1zoom != $::XRlab::x2zoom} {
    set radius [expr {sqrt(($::XRlab::y1zoom-$::XRlab::y2zoom)**2 + ($::XRlab::x1zoom != $::XRlab::x2zoom)**2)/2}]
    set ym [expr {int(($::XRlab::y1zoom+$::XRlab::y2zoom)/2.0)}]
    set xm [expr {int(($::XRlab::x1zoom+$::XRlab::x2zoom)/2.0)}]
    
    set zmin [$bfl get $ym $xm]
    set zmax $zmin

    # reset mask 
    $mask = 0
    
    switch $selectionmode {
      circle {
       for {set y 1} {$y<511} {incr y} {
         for {set x 1} {$x<511} {incr x} {
	   if {($x-$xm)**2+($y-$ym)**2 <$radius**2} {
	     $mask set $y $x 1
	   }  
	 }
       }
      } 

      rectangle {
       set x1 [expr {max($XRlab::x1zoom, 1)}]
       set x2 [expr {min($XRlab::x2zoom, 510)}]
       set y1 [expr {max($XRlab::y1zoom, 1)}]
       set y2 [expr {min($XRlab::y2zoom, 510)}]

       for {set y $y1} {$y<$y2} {incr y} {
         for {set x $x1} {$x<$x2} {incr x} {
	   $mask set $y $x 1
	 }
       }	 
      }
      
      mask {
        # copy the provided mask
	# do not use $mask = $selectionmask
        # because we want mask to be a normalized bild16
        MAP_MASK $selectionmask {
	  $mask set $y $x 1
	}
      }	

      default {error "Invalid selectionmode"}
    }  
    
    # to be sure to not touch the border
    # reset a 2 pixel border around the masks
    for {set c 0} {$c<512} {incr c} {
      #left
      $mask set 0 $c 0
      $mask set 1 $c 0
      # top
      $mask set $c 0 0
      $mask set $c 1 0
      #right
      $mask set 510 $c 0
      $mask set 511 $c 0
      #bottom
      $mask set $c 510 0
      $mask set $c 511 0
    }

    MAP_MASK $mask {
      set z [$bfl get $y $x]
      if {$z>$zmax} {set zmax $z}
      if {$z<$zmin} {set zmin $z}
    }

    if {$zmin==$zmax} {
      set zmin [expr {$zmin*0.95}]
      set zmax [expr {$zmax*1.05}]
    }  

    set zm [expr {($zmin+$zmax)/2}]

    glBindTexture GL_TEXTURE_1D $colortexture
    glTexEnvf GL_TEXTURE_ENV GL_TEXTURE_ENV_MODE $::GL_MODULATE
    set texturescale [expr {1.0/($zmax-$zmin)}]
    set textureoffset [expr {-($zmin-$zm)*$texturescale}]
    set ::currentCoeff   [list 0.0 0.0 $texturescale $textureoffset]
    glTexGeni GL_S GL_TEXTURE_GEN_MODE $::GL_OBJECT_LINEAR
    glTexGenfv GL_S $::GL_OBJECT_PLANE $::currentCoeff

    glEnable GL_TEXTURE_GEN_S
    glEnable GL_TEXTURE_1D
    glBegin GL_QUADS
      if {$::XRlab::y1zoom!=$::XRlab::y2zoom && $::XRlab::x1zoom != $::XRlab::x2zoom} {
        set radius [expr {sqrt(($::XRlab::y1zoom-$::XRlab::y2zoom)**2 + ($::XRlab::x1zoom != $::XRlab::x2zoom)**2)/2}]
        set ym [expr {($::XRlab::y1zoom+$::XRlab::y2zoom)/2.0}]
        set xm [expr {($::XRlab::x1zoom+$::XRlab::x2zoom)/2.0}]
        set zm [expr {($zmin+$zmax)/2}]
  
        MAP_MASK $mask {
	  # exclude also points, where one of the vertices 
	  # at 1,0 1,1 or 0,1 is outside the masked range
	  set normalvecs {}
	  set vertexvecs {}
	  set outside 0
          foreach {ex ey} {0 0   1 0   1 1    0 1} {
            set nx [expr {$x+$ex}]
            set ny [expr {$y+$ey}]
	    if {[$mask get $ny $nx]==0} {
	      # exclude this quad
	      set outside 1
	      break
	    }
            set z [$bfl get $ny $nx]
      
            set xd [expr {($nx-$xm)*$px}]
            set yd [expr {($ym-$ny)*$px}]
            set zd [expr {($z-$zm)}]
        
	    # calculate the normal vector - get the central differences in 
	    # x- and y-direction
	    set normal_x [expr {[$bfl get $ny [expr {$nx+1}]]-[$bfl get $ny [expr {$nx-1}]]}]

	    set normal_y [expr {[$bfl get [expr {$ny+1}] $nx]-[$bfl get [expr {$ny-1}] $nx]}]
            set normal_z [expr {2*$px}]

	    set norm [expr {sqrt($normal_x**2+$normal_y**2+$normal_z**2)}]
            
	    lappend normalvecs [expr {-$normal_x/$norm}] [expr {$normal_y/$norm}] [expr {$normal_z/$norm}]
	    lappend vertexvecs $xd $yd $zd
          }

	  if {!$outside} {
	    foreach {nx ny nz} $normalvecs {xd yd zd} $vertexvecs {
	      glNormal3f $nx $ny $nz
	      glVertex3f $xd $yd $zd
	    }
	  }

       }
     }   
   glEnd
   glDisable GL_TEXTURE_1D
   glDisable GL_TEXTURE_GEN_S

 } else {
   error "Bad selection"
 }
}

proc make_colorbar {} {
  variable zmin
  variable zmax
  variable colortexture
  variable contour_dist
  variable contour_halfwidth

  # compute boundaries for the contour lines in z space
  set contour_bounds {}
  set hdiff [expr {$zmax-$zmin}]
  for {set cnum [expr {int(ceil($zmin/$contour_dist))}]} {$cnum*$contour_dist<$zmax} {incr cnum} {
    lappend contour_bounds [expr {$cnum*$contour_dist-$contour_halfwidth}] [expr {$cnum*$contour_dist+$contour_halfwidth}] 
  }

  # append "infinite" bounds that are for sure larger 
  # 
  lappend contour_bounds [expr {$zmax*2+$contour_halfwidth}] [expr {$zmax*2+3*$contour_halfwidth}]

  glBindTexture GL_TEXTURE_1D $colortexture

    glTexParameteri GL_TEXTURE_1D GL_TEXTURE_MAG_FILTER $::GL_LINEAR
    glTexParameteri GL_TEXTURE_1D GL_TEXTURE_MIN_FILTER $::GL_LINEAR_MIPMAP_LINEAR

    glTexParameteri GL_TEXTURE_1D GL_TEXTURE_WRAP_S $::GL_REPEAT
  
  
    set length 2048
    set teximage [tcl3dVector GLubyte [expr $length*4]]

    set contour_bounds [lassign $contour_bounds clowbound chighbound]
    
    for { set i 0 } { $i < $length } { incr i } {
      set cb [expr {double($i)/$length}]
      # compute corresponding height
      set z [expr {$hdiff*$cb+$zmin}]
      
      # compute color according to colormap
      set red [expr {int([red $cb]*255)}]
      set green [expr {int([green $cb]*255)}]
      set blue [expr {int([blue $cb]*255)}]
      
      # if we are inside a contour line, change to black
      if {$z>$clowbound} { lassign {0 0 0} red green blue }

      if {$z>$chighbound} {
       # if we cross the upper bound of the contour line, shift
       set contour_bounds [lassign $contour_bounds clowbound chighbound]
      }

      $teximage set [expr {($i)*4 + 0}] $red
      $teximage set [expr {($i)*4 + 1}] $green
      $teximage set [expr {($i)*4 + 2}] $blue
      $teximage set [expr {($i)*4 + 3}] 0xff
    }
  

     gluBuild1DMipmaps GL_TEXTURE_1D 4 $length \
                        GL_RGBA GL_UNSIGNED_BYTE $teximage
      
     $teximage delete
}

proc PovRayExportButton {} {
  # ask the user for options
  variable viewlist 
  variable px
  variable xm
  variable ym
  set what View
  if {[llength $viewlist]>0} {
    foreach {ftime camx camy camz lx ly lz angle} $viewlist {
      lassign [RotateVector [list $camx $camy $camz] {0 0 1} [expr {-$angle}]] camobjx camobjy
      set myx [expr {$camobjx/$px+$xm}]    
      set myy [expr {$ym-$camobjy/$px}] 
      
      lappend camlist $myx $myy
    }
    $::XRlab::can create line $camlist -fill yellow -tag flight -width 3
    $::XRlab::can create line $camlist -fill black -tag flight -width 1

    set ans [tk_messageBox -type yesno -message "There exists a recorded flight (see topview). Do you want to export the animation to PovRay? If no, only the current view is exported." -icon question]
    if {$ans=="yes"} {
      set what Flight
    }
    $::XRlab::can delete flight
  }
  
  # Open a dialog window and prompt for file name
  set outputname [tk_getSaveFile -filetypes {{{PovRay} {.pov}}} -defaultextension .pov]
  if {$outputname==""} {return}
  set outputbase [file rootname $outputname]
 
  PovRayExport $what $outputbase
}

proc PovRayExport {what outputbase} {
  # export the visualization as a PovRay scene
  # what=View   only current View
  # what=Flight animation of the content of viewlist
  variable ns
  variable scriptdir
  variable toglwin
  variable viewlist
  variable objangle
  variable bulletlist


  variable alpha_GL
  variable camx
  variable camy
  variable camz
  variable centerx
  variable centery
  variable centerz

  variable bfl
  variable mask
  variable px

  variable xm
  variable ym
  variable radius
  variable zmin
  variable zmax
  variable zm

  
  bildfloat povrayfloat $bfl
  bild16 povray16 
  povrayfloat -= $zmin
  set heightdiff [expr {$zmax-$zmin}]
  povray16 fromfloat povrayfloat [expr {($heightdiff)*65535./65534}]
  povray16 += 1
 
#  MAP_UNCIRCLE_XY $xm $ym $radius {
#    povray16 set $y $x 0
#  }

  povray16 *= $mask

  # generate height field
  povray16 write_pgm16 "$outputbase.pgm"

  # generate povray scene
  set fd [open [file join $scriptdir povexport.povtemplate] r]
  set povtemplate [read $fd]
  close $fd

  # generate povray ini file 
  set fd [open [file join $scriptdir povexport.initemplate] r]
  set initemplate [read $fd]
  close $fd

  
  switch $what {
    View {
      set camposlist "0, <$camx,$camz,$camy>\n"
      set lookatlist "0, <$centerx,$centerz,$centery>\n"
      set anglelist "0, <0,$objangle,0>\n"
    }
    
    Flight {
      set mindist 0.1
      set minangle 3
      set viewlist [lassign $viewlist stime lcx lcy lcz llx lly llz langle]
      append camposlist "0, <$lcx,$lcz,$lcy>\n"
      append lookatlist "0, <$llx,$llz,$lly>\n"
      append anglelist  "0, <0,$langle,0>\n"
    
      set finaltime [lindex $viewlist end-7]
      # generate spline coordinates
      foreach {ftime camx camy camz lx ly lz angle} $viewlist {
        set dist [expr {sqrt(($camx-$lcx)**2+($camy-$lcy)**2)}]
        if {$dist>$mindist || abs($angle-$langle)>$minangle || $ftime==$finaltime} {
          set povtime [expr {double($ftime-$stime)/($finaltime-$stime)}]
          append camposlist "$povtime, <$camx,$camz,$camy>\n"
          append lookatlist "$povtime, <$lx,$lz,$ly>\n"
          append anglelist  "$povtime, <0,$angle,0>\n"
          set lcx $camx
          set lcy $camy
          set lcz $camz
	  set langle $angle
        }
      }
    }

  }  


  set width [winfo width $toglwin]
  set height [winfo height $toglwin]
  
  set color_map {} 
  variable colormap_quantum 0.02
  variable contour_dist 
  variable contour_halfwidth

  # compute boundaries for the contour lines in z space
  set contour_bounds {}
  for {set cnum [expr {int(ceil($zmin/$contour_dist))}]} {$cnum*$contour_dist<$zmax} {incr cnum} {
    lappend contour_bounds [expr {$cnum*$contour_dist-$contour_halfwidth}] [expr {$cnum*$contour_dist+$contour_halfwidth}] 
  }

  # append "infinite" bounds that are for sure larger 
  # 
  lappend contour_bounds [expr {$zmax*2+$contour_halfwidth}] [expr {$zmax*2+3*$contour_halfwidth}]

  # if {$colormap_quantum>$contour_halfwidth/$heightdiff} {error "Colormap Quantum too large"}
  
  set halfstep [expr {$colormap_quantum/2}]
  
  set contour_bounds [lassign $contour_bounds clowbound chighbound]

  for {set cb 0} {$cb<=1.0+$halfstep} {set cb [expr {$cb+$colormap_quantum}]} {
    # compute color code for the color code cb
    set red [red $cb]
    set green [green $cb]
    set blue [blue $cb]
    # compute corresponding height
    set z [expr {$heightdiff*$cb+$zmin}]
    # whenever z crosses a contour line
    # append color-black black-color at both ends
    if {$z>$clowbound} {
       set cblow  [expr {($clowbound-$zmin)/$heightdiff}]
       set cbhigh [expr {($chighbound-$zmin)/$heightdiff}]
       append color_map "\[$cblow color rgb <$red,$green,$blue>\]\n"
       append color_map "\[$cblow color Black\]\n"
       append color_map "\[$cbhigh color Black\]\n"
       append color_map "\[$cbhigh color rgb <$red,$green,$blue>\]\n"
       # pop the next contour boundaries from the stack
       set contour_bounds [lassign $contour_bounds clowbound chighbound]
       set cb $cbhigh
    } else {
       append color_map "\[$cb color rgb <$red,$green,$blue>\]\n"
    }   
  }

  # bullets
  set bulletnumber 0
  if {[llength $bulletlist]!=0} {
    set bulletrlist [lassign $bulletlist bulx buly bulz]
    set bulletpositionlist "<$bulx,$bulz,$buly>"
    incr bulletnumber
    foreach {bulx buly bulz} $bulletrlist {
      append bulletpositionlist ",<$bulx,$bulz,$buly>"
      incr bulletnumber
    }
  } else {
    set bulletpositionlist {<0,0,0>}
  }
  set povcode [subst -nocommands -nobackslashes $povtemplate]

  set fd [open "$outputbase.pov" w]
  puts $fd $povcode
  close $fd

  set inicode [subst -nocommands -nobackslashes $initemplate]

  set fd [open "$outputbase.ini" w]
  puts $fd $inicode
  close $fd


}  


proc qadd {q p} {
  # add two quaternions
  lassign $q a1 b1 c1 d1
  lassign $p a2 b2 c2 d2

  list [expr {$a1+$a2}] [expr {$b1+$b2}] [expr {$c1+$c2}] [expr {$d1+$d2}] 
}

proc qquer {q} {
  # conjugate a quaternion
  lassign $q a b c d
  list $a [expr {-$b}] [expr {-$c}] [expr {-$d}] 
}

proc qmul {q p} {
  # multiply two quaternions
  lassign $q a1 b1 c1 d1
  lassign $p a2 b2 c2 d2
  
  list \
    [expr {$a1*$a2-$b1*$b2-$c1*$c2-$d1*$d2}] \
    [expr {$a1*$b2+$b1*$a2+$c1*$d2-$d1*$c2}] \
    [expr {$a1*$c2-$b1*$d2+$c1*$a2+$d1*$b2}] \
    [expr {$a1*$d2+$b1*$c2-$c1*$b2+$d1*$a2}] 
}


proc RotateVector {v n alpha} {
  # rotate vector v around axis n by angle alpha
  # using quaternion algebra
  # n must be a unit vector
  lassign $n nx ny nz
  set theta [expr {$alpha*3.14159265358979323/180/2}]
  set q [list [expr {cos($theta)}] [expr {sin($theta)*$nx}] [expr {sin($theta)*$ny}] [expr {sin($theta)*$nz}]] 
  # rotation formula is q v q*
  set p [linsert $v 0 0]
  set qrot [qmul $q [qmul $p [qquer $q]]]

  lrange $qrot 1 end
}

#  
# dialog box for options
#
# name description default min max increment
    set optionvars {
      contour_dist 		"Contour distance"		1.0	0.1	10	0.1
      contour_halfwidth		"Contour half width"		0.03	0.005 	1	0.005
      runspeed			"Quake speed"			10.0	0	100	5
      camx			"Camera x"			0.0	-2000	2000	1
      camy			"Camera y"			-50.0	-2000	2000	1
      camz			"Camera z"			50.0	0	2000	1
      tiltangle			"Latitude"			45.0	-90	90	1
      turnangle			"Altitude"			0.0	-360	360	1
      objangle			"Object rotation"		0.0	-360	360	1
      alpha_GL			"Perspective angle (y)"		38.0	10.0	90.0	1
    } 


proc optiondialog {} {
    variable toglwin
    variable ns
    variable optionvars
    ::iwidgets::dialog .optiond -modality application -title "XRay3D options"
    set cs [.optiond childsite]

    foreach {var desc def min max inc} $optionvars {
      ttk::label $cs.${var}_l -text $desc
      #entry $cs.$var -textvariable ${ns}::$var
      spinbox $cs.$var -textvariable ${ns}::$var -command ${ns}::redraw -from $min -to $max -increment $inc 
      bind $cs.$var <FocusOut> [list ${ns}::rangevalidate $min $max $def ${ns}::$var]
      grid  $cs.${var}_l   $cs.$var -sticky nsew
      set save($var) [set ${ns}::$var]
    }

    # selectionmode
    set selmode [ttk::frame $cs.selmode]
    ttk::label $selmode.l -text "Selection"
    ttk::radiobutton $selmode.circle -text "Circular" -variable ${ns}::selectionmode -value circle -command ${ns}::enablemaskentry
    ttk::radiobutton $selmode.rectangle -text "Rectangular" -variable ${ns}::selectionmode -value rectangle -command ${ns}::enablemaskentry
    ttk::radiobutton $selmode.mask -text "Mask" -variable ${ns}::selectionmode -value mask -command ${ns}::enablemaskentry
    pack $selmode.l $selmode.circle $selmode.rectangle $selmode.mask -side left
    variable maskentry [entry $cs.selectmask -textvariable ${ns}::selectionmask]
    enablemaskentry
    
    grid $cs.selmode $cs.selectmask
    .optiond add Reset -text "Load Defaults" -command ${ns}::loaddefaults
    .optiond hide Apply
    .optiond hide Help
    
    # run the dialog
    set result [.optiond activate]
 
    ::itcl::delete object .optiond 
    # don't apply if user hits Cancel
    if {!$result} {
      # undo
      foreach {var desc def min max inc} $optionvars {
        set ${ns}::$var $save($var)
      }
    } 
    
    # apply changes   
    redraw
 
  }

proc redraw {} {
  variable toglwin
  set w [winfo width $toglwin]
  set h [winfo height $toglwin]
  tclReshapeFunc $toglwin $w $h
  make_colorbar
  updateView
}  

proc loaddefaults {} {
  variable ns
  variable optionvars
  foreach {var desc def min max inc} $optionvars {
    set ${ns}::$var $def
  }
  redraw
}

proc rangevalidate {min max def var} {
  # first make sure that value can be converted
  # into a double
  set value [set $var]
  if {![scan $value %f floatval]} {
    set $var $def
  } else {  
    # it is a double (prefix)
    # value outside the allowed range ?
    if {$floatval<$min || $floatval>$max} {
       set $var $def
    } else {
      set $var $floatval
    }  
  }  
  redraw 
}

proc enablemaskentry {} {
  variable maskentry
  variable selectionmode
  if {$selectionmode=="mask"} {
    $maskentry configure -state normal
  } else {
    $maskentry configure -state disabled
  }  
}

proc ::MAP_MASK {mask command} {
  set MAP_up [list upvar 1]
  set locvars [uplevel 1 {info vars}]
  foreach locvar $locvars {
    lappend MAP_up $locvar $locvar
  }

  if {[llength $locvars]==0} {set MAP_up ""}
  
  proc mapbody {} "$MAP_up
    for {set y 0} {\$y<512} {incr y} {
     for {set x 0} {\$x<512} {incr x} {
       if {\[$mask get \$y \$x\]>0} {$command}
     }
    }"
  uplevel 1 mapbody
}

